from string import Template
import logging
import json
import pprint
import jsondiff
import requests
import os
import jwt
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter
import datetime


LEMANS_RESOURCE_SERVICE = (
    os.getenv("LEMANS_RESOURCE_SERVICE") or "lemans-resources-service"
)
LEMANS_PORT = os.getenv("LEMANS_PORT") or "80"

KAFKA_SERVICE = os.getenv("DEFAULT_KAFKA") or "ops-kafka-client"
KAFKA_NAMESPACE = os.getenv("DEFAULT_KAFKA_NAMESPACE")
KAFKA_SSL_TO_PLAINTEXT_PORT = "9093"
KAFKA_SSL_PORT = "9092"
KAFKA_PLAINTEXT_PORT = "9092"
KAFKA_KEY_SERIALIZER = "org.apache.kafka.common.serialization.StringSerializer"
KAFKA_VALUE_SERIALIZER = "org.apache.kafka.common.serialization.StringSerializer"
KAFKA_KEY_DESERIALIZER = "org.apache.kafka.common.serialization.StringDeserializer"
KAFKA_VALUE_DESERIALIZER = "org.apache.kafka.common.serialization.StringDeserializer"
KAFKA_ACKS = "1"
KAFKA_RETRIES = "2"
KAFKA_BATCH_SIZE = "16384"
KAFKA_MAX_BLOCK_MS = "6000"
KAFKA_MAX_REQUEST_SIZE = "5242880"
KAFKA_LINGER_MS = "5"
KAFKA_BUFFER_MEMORY = "33554432"
KAFKA_SECURITY_PROTOCOL = os.getenv("KAFKA_SECURITY_PROTOCOL") or "SSL"
KAFKA_SSL_KEYSTORE_PASS = "VMware1234"
KAFKA_SSL_KEY_PASS = "VMware1234"
KAFKA_SSL_TRUSTSTORE_PASS = "VMware1234"
KAFKA_PARTITIONER_CLASS = "com.vmware.lemans.receivers.kafka.RoundRobinPartitioner"
KAFKA_NUMBER_OF_PARTITIONS = int(os.getenv("KAFKA_PARTITIONS")) or 1
KAFKA_REPLICAS = int(os.getenv("KAFKA_REPLICAS")) or 1
KAFKA_RETENTION_PERIOD = int(os.getenv("KAFKA_RETENTION_PERIOD")) or 7200000
KAFKA_CLEANUP_POLICY = os.getenv("KAFKA_CLEANUP_POLICY") or "delete"
true = "true"
false = "false"
SM_CONTEXT = os.getenv("SM_CONTEXT") or "cloud"
SM_ORGID = os.getenv("SM_ORGID") or ""
print("SM_CONTEXT is :", SM_CONTEXT)
if SM_CONTEXT == "sm-hub":
    # port needs to be rechecked to point to 8000
    LEMANS_PORT = "8000"
    SM_EBALED = True
    orgID = SM_ORGID
    print("defaultOrgId is : ", orgID)
    if not orgID:
        print("defaultOrgId not found in the file")

    now = datetime.datetime.utcnow()

    private_key_pem = ""
    file_exist = os.path.isfile("/run/auth/ssl/private.pem")
    if file_exist:
        with open("/run/auth/ssl/private.pem", "r") as f:
            private_key_pem = f.read()
    else:
        print("private_key.pem file not found in the pod")

    print("reading the file from /run/auth/ssl/private.pem :", private_key_pem)

    payload = {
        "iat": now,
        "exp": now + datetime.timedelta(hours=1),
        "org_id": orgID,
        "context_name": orgID,
        "sub": "1234567890",
        "issuer": "le-mans-gateway",
    }
    # Algorithm to use for signing
    algorithm = "RS256"

    # Create the token
    jwttoken = jwt.encode(payload, private_key_pem, algorithm=algorithm)

    print("JWT Token:", jwttoken)
    logging.info("%s JWT Token:", jwttoken)
    jwttoken = "Bearer " + jwttoken
    print("Bearer JWT Token:", jwttoken)
    logging.info("Stream creations and updations are done in SM world")
    requests.get(
        "http://"
        + LEMANS_RESOURCE_SERVICE
        + "."
        + KAFKA_NAMESPACE
        + ".svc.cluster.local:"
        + LEMANS_PORT
    )
else:
    SM_EBALED = False

logging.getLogger().setLevel(logging.DEBUG)


def create(namespace, csp_uri):

    lemans_hostname = ""
    lemans_service_exists = check_if_service_exists(
        namespace, LEMANS_RESOURCE_SERVICE, LEMANS_PORT
    )
    if not lemans_service_exists:
        logging.info(
            "Could not find the LeMans service. Skip registration with LeMans."
        )
        return None
    else:
        lemans_hostname = get_service_hostname(namespace, LEMANS_RESOURCE_SERVICE)
    kafka_hostname = get_service_hostname(KAFKA_NAMESPACE, KAFKA_SERVICE)

    lemans_resources_adapter = _get_session_with_lemans_proxy_adapter(csp_uri)
    return LemansRegistrar(
        lemans_resources_adapter, lemans_hostname, kafka_hostname, namespace
    )


def _get_session_with_lemans_proxy_adapter(csp_uri):
    with requests.Session() as requests_session:

        def get_token():
            return get_csp_access_token(csp_uri)

    adapter = AuthenticatingAdapter(get_token)
    return session(adapter)


def get_csp_access_token(csp_uri):

    payload = "grant_type=client_credentials"
    csp_oauth_uri = csp_uri + "/csp/gateway/am/api/auth/authorize"
    client_id = os.getenv("CLIENT_ID")
    client_secret = os.getenv("CLIENT_SECRET")
    response = requests.post(
        csp_oauth_uri,
        data=payload,
        auth=HTTPBasicAuth(client_id, client_secret),
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )
    response.raise_for_status()

    data = json.loads(response.text)
    csp_access_token = data["access_token"]
    if not csp_access_token:
        raise Exception("csp access token for %s is empty" + csp_oauth_uri)
        return False
    logging.info("CSP Access token is %s", csp_access_token)
    return csp_access_token


def check_if_service_exists(namespace, service_name, service_port):
    try:
        logging.info(
            "checking if the service exists : "
            + "http://"
            + service_name
            + "."
            + namespace
            + ".svc.cluster.local:"
            + service_port
        )
        response = requests.get(
            "http://"
            + service_name
            + "."
            + namespace
            + ".svc.cluster.local:"
            + service_port
        )
        return True
    except Exception:
        return False


class LemansRegistrar:
    def __init__(
        self, lemans_resources_adapter, lemans_hostname, kafka_hostname, namespace
    ):
        self.lemans_resources_adapter = lemans_resources_adapter
        self.lemans_hostname = lemans_hostname
        self.kafka_hostname = kafka_hostname
        self.namespace = namespace

    def register_service_to_lemans(self, stream_name="", receivers=[]):
        kafka_receiver = {}
        receiver = {}
        egress_receivers = []
        kafka_config_exists = False
        sample_config_exists = False
        sample_receiver = {}
        logging.info("raw receivers are : {0} ".format(receivers))
        for receiver_config in receivers:
            # logging.info("receivers configs are : {0} ".format(receiver_config))
            if "kafka" in receiver_config.values():
                kafka_config_exists = True
                kafka_receiver = receiver_config
                logging.info(
                    "kafka_receiver configs are : {0} ".format(receiver_config)
                )
            elif "sample" in receiver_config.values():
                sample_config_exists = True
                sample_receiver = receiver_config
                logging.info(
                    "Sample_receiver configs are : {0} ".format(sample_receiver)
                )
            else:
                receiver = receiver_config
                logging.info("egress_receivers configs are : {0} ".format(receiver))
                egress_receivers.append(receiver_config)
        logging.info("egress_receivers list : {0} ".format(egress_receivers))

        sample_receiver_name = ""
        if sample_config_exists:
            sample_receiver_name = sample_receiver["name"]
            sample_receiver_options = sample_receiver["sample_receiver_options"]
            if self.receiver_starter_exists(sample_receiver_name):
                self.update_sample_receiver_starter(
                    receiver_starter_name=sample_receiver_name,
                    stream_name=stream_name,
                    sample_receiver_options=sample_receiver_options,
                )
            else:
                self.create_sample_receiver_starter(
                    receiver_starter_name=sample_receiver_name,
                    stream_name=stream_name,
                    sample_receiver_options=sample_receiver_options,
                )
            if not self.receiver_exists(sample_receiver_name):
                self.create_receiver(
                    sample_receiver,
                    "/le-mans/sample-receivers" + "/" + sample_receiver_name,
                )
            else:
                self.update_receiver(
                    sample_receiver,
                    "/le-mans/sample-receivers" + "/" + sample_receiver_name,
                )
            self.multiplexer_operations(sample_receiver)

        if len(egress_receivers) == 1:
            logging.info("we have only one egress_receiver")
            kafka_receiver_name = ""
            if kafka_config_exists:
                kafka_k8s_svc = KAFKA_SERVICE
                kafka_k8s_ns = KAFKA_NAMESPACE
                kafka_tls_protocol = KAFKA_SECURITY_PROTOCOL
                if "kafka_k8s_svc" in kafka_receiver:
                    kafka_k8s_svc = kafka_receiver["kafka_k8s_svc"]
                if "kafka_k8s_ns" in kafka_receiver:
                    kafka_k8s_ns = kafka_receiver["kafka_k8s_ns"]
                if "kafka_tls_protocol" in kafka_receiver:
                    kafka_tls_protocol = kafka_receiver["kafka_tls_protocol"]
                kafka_receiver_name = kafka_receiver["name"]
                kafka_producer_name = ""
                if "kafka_producer_name" in kafka_receiver:
                    kafka_producer_name = kafka_receiver["kafka_producer_name"]
                kafka_producer_options = ""
                if "kafka_producer_options" in kafka_receiver:
                    kafka_producer_options = kafka_receiver["kafka_producer_options"]
                kafka_consumer_name = kafka_receiver["kafka_consumer_name"]
                kafka_consumer_options = kafka_receiver["kafka_consumer_options"]
                if "group.id" not in kafka_consumer_options:
                    logging.error(
                        "Please specify group.id in kafka_consumer_options in lemans hook configuration for %s",
                        kafka_consumer_name,
                    )
                    raise Exception
                lemans_multiplexer_name = ""
                if "lemans_multiplexer_name" in kafka_receiver:
                    lemans_multiplexer_name = kafka_receiver["lemans_multiplexer_name"]
                self.multiplexer_operations(kafka_receiver)
                consumer_only = False
                if "consumer_only" in kafka_receiver:
                    consumer_only = kafka_receiver["consumer_only"]

                no_of_consumers = 0
                if "no_of_consumers" in kafka_receiver:
                    no_of_consumers = int(kafka_receiver["no_of_consumers"])

                if "start_json_state_producer_options" in kafka_receiver:
                    start_json_state_producer_options = kafka_receiver[
                        "start_json_state_producer_options"
                    ]
                else:
                    start_json_state_producer_options = ""

                if "start_json_state_consumer_options" in kafka_receiver:
                    start_json_state_consumer_options = kafka_receiver[
                        "start_json_state_consumer_options"
                    ]
                else:
                    start_json_state_consumer_options = ""

                if kafka_producer_name and not consumer_only:
                    if self.receiver_starter_exists(kafka_producer_name):
                        self.update_kafka_receiver_starter(
                            kafka_producer_name,
                            "producer",
                            start_json_state_options=start_json_state_producer_options,
                            kafka_options=kafka_producer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
                    else:
                        self.create_kafka_receiver_starter(
                            kafka_producer_name,
                            "producer",
                            start_json_state_options=start_json_state_producer_options,
                            kafka_options=kafka_producer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
                if not consumer_only:
                    if not self.receiver_exists(kafka_receiver_name):
                        if lemans_multiplexer_name:
                            self.create_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/multiplexer"
                                + "/"
                                + lemans_multiplexer_name,
                            )
                        else:
                            self.create_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/kafka-producer"
                                + "/"
                                + kafka_producer_name,
                            )
                    else:
                        if lemans_multiplexer_name:
                            self.update_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/multiplexer"
                                + "/"
                                + lemans_multiplexer_name,
                            )
                        else:
                            self.update_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/kafka-producer"
                                + "/"
                                + kafka_producer_name,
                            )
                if kafka_consumer_name:
                    if no_of_consumers > 0:
                        for x in range(1, no_of_consumers + 1):
                            kafka_consumer_name_part = (
                                kafka_consumer_name + "-" + str(x)
                            )
                            if self.receiver_starter_exists(kafka_consumer_name_part):
                                self.update_kafka_receiver_starter(
                                    kafka_consumer_name_part,
                                    "consumer",
                                    start_json_state_options=start_json_state_consumer_options,
                                    kafka_options=kafka_consumer_options,
                                    stream_name=stream_name,
                                    kafka_producer_name=kafka_producer_name,
                                    kafka_receiver_name=kafka_receiver_name,
                                    receiver_name=receiver["name"],
                                    kafka_k8s_svc=kafka_k8s_svc,
                                    kafka_k8s_ns=kafka_k8s_ns,
                                    kafka_tls_protocol=kafka_tls_protocol,
                                )
                            else:
                                self.create_kafka_receiver_starter(
                                    kafka_consumer_name_part,
                                    "consumer",
                                    start_json_state_options=start_json_state_consumer_options,
                                    kafka_options=kafka_consumer_options,
                                    stream_name=stream_name,
                                    kafka_producer_name=kafka_producer_name,
                                    kafka_receiver_name=kafka_receiver_name,
                                    receiver_name=receiver["name"],
                                    kafka_k8s_svc=kafka_k8s_svc,
                                    kafka_k8s_ns=kafka_k8s_ns,
                                    kafka_tls_protocol=kafka_tls_protocol,
                                )

                if kafka_consumer_name:
                    if self.receiver_starter_exists(kafka_consumer_name):
                        self.update_kafka_receiver_starter(
                            kafka_consumer_name,
                            "consumer",
                            start_json_state_options=start_json_state_consumer_options,
                            kafka_options=kafka_consumer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            receiver_name=receiver["name"],
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
                    else:
                        self.create_kafka_receiver_starter(
                            kafka_consumer_name,
                            "consumer",
                            start_json_state_options=start_json_state_consumer_options,
                            kafka_options=kafka_consumer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            receiver_name=receiver["name"],
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
            if receiver:
                self.create_egress_receiver_with_routes_and_consumer(
                    stream_name, receiver, kafka_receiver
                )
        else:
            if not kafka_config_exists:
                logging.info(
                    "we have multiple egress_receiver without type: kafka config"
                )
                for receiver in egress_receivers:
                    self.create_egress_receiver_with_routes_and_consumer(
                        stream_name, receiver, kafka_receiver
                    )

            kafka_receiver_name = ""
            logging.info("we have multiple egress_receivers with kafka receiver")
            if kafka_config_exists:
                kafka_k8s_svc = KAFKA_SERVICE
                kafka_tls_protocol = KAFKA_SECURITY_PROTOCOL
                kafka_k8s_ns = KAFKA_NAMESPACE
                if "kafka_k8s_svc" in kafka_receiver:
                    kafka_k8s_svc = kafka_receiver["kafka_k8s_svc"]
                if "kafka_k8s_ns" in kafka_receiver:
                    kafka_k8s_ns = kafka_receiver["kafka_k8s_ns"]
                if "kafka_tls_protocol" in kafka_receiver:
                    kafka_tls_protocol = kafka_receiver["kafka_tls_protocol"]

                consumer_only = False
                if "consumer_only" in kafka_receiver:
                    consumer_only = kafka_receiver["consumer_only"]

                kafka_receiver_name = kafka_receiver["name"]
                kafka_producer_name = ""
                if "kafka_producer_name" in kafka_receiver:
                    kafka_producer_name = kafka_receiver["kafka_producer_name"]
                kafka_producer_options = ""
                if "kafka_producer_options" in kafka_receiver:
                    kafka_producer_options = kafka_receiver["kafka_producer_options"]
                if "start_json_state_producer_options" in kafka_receiver:
                    start_json_state_producer_options = kafka_receiver[
                        "start_json_state_producer_options"
                    ]
                else:
                    start_json_state_producer_options = ""
                lemans_multiplexer_name = ""
                if "lemans_multiplexer_name" in kafka_receiver:
                    lemans_multiplexer_name = kafka_receiver["lemans_multiplexer_name"]
                self.multiplexer_operations(kafka_receiver)
                for receiver in egress_receivers:
                    if "kafka_consumer_name" in receiver.keys():
                        kafka_consumer_name = receiver["kafka_consumer_name"]
                        kafka_consumer_options = receiver["kafka_consumer_options"]
                        if "group.id" not in kafka_consumer_options:
                            logging.error(
                                "Please specify group.id in kafka_consumer_options in lemans hook configuration for %s",
                                kafka_consumer_name,
                            )
                            raise Exception

                if kafka_producer_name and not consumer_only:
                    if self.receiver_starter_exists(kafka_producer_name):
                        self.update_kafka_receiver_starter(
                            kafka_producer_name,
                            "producer",
                            start_json_state_options=start_json_state_producer_options,
                            kafka_options=kafka_producer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
                    else:
                        self.create_kafka_receiver_starter(
                            kafka_producer_name,
                            "producer",
                            start_json_state_options=start_json_state_producer_options,
                            kafka_options=kafka_producer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
                if not consumer_only:
                    if not self.receiver_exists(kafka_receiver_name):
                        if lemans_multiplexer_name:
                            self.create_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/multiplexer"
                                + "/"
                                + lemans_multiplexer_name,
                            )
                        else:
                            self.create_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/kafka-producer"
                                + "/"
                                + kafka_producer_name,
                            )
                    else:
                        if lemans_multiplexer_name:
                            self.update_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/multiplexer"
                                + "/"
                                + lemans_multiplexer_name,
                            )
                        else:
                            self.update_receiver(
                                kafka_receiver,
                                "/le-mans/receivers/kafka-producer"
                                + "/"
                                + kafka_producer_name,
                            )
                for receiver in egress_receivers:
                    if not self.receiver_exists(receiver["name"]):
                        self.create_receiver(receiver, receiver["receiver_address"])
                    else:
                        self.update_receiver(receiver, receiver["receiver_address"])

                    if "kafka_consumer_name" in receiver.keys():
                        no_of_consumers = 0
                        if "no_of_consumers" in receiver:
                            no_of_consumers = int(receiver["no_of_consumers"])

                        kafka_consumer_name = receiver["kafka_consumer_name"]
                        if "start_json_state_consumer_options" in receiver:
                            start_json_state_consumer_options = receiver[
                                "start_json_state_consumer_options"
                            ]
                        else:
                            start_json_state_consumer_options = ""

                        kafka_consumer_options = receiver["kafka_consumer_options"]

                        if no_of_consumers > 0:
                            for x in range(1, no_of_consumers + 1):
                                kafka_consumer_name_part = (
                                    kafka_consumer_name + "-" + str(x)
                                )
                                if self.receiver_starter_exists(
                                    kafka_consumer_name_part
                                ):
                                    self.update_kafka_receiver_starter(
                                        kafka_consumer_name_part,
                                        "consumer",
                                        start_json_state_options=start_json_state_consumer_options,
                                        kafka_options=kafka_consumer_options,
                                        stream_name=stream_name,
                                        kafka_producer_name=kafka_producer_name,
                                        kafka_receiver_name=kafka_receiver_name,
                                        receiver_name=receiver["name"],
                                        kafka_k8s_svc=kafka_k8s_svc,
                                        kafka_k8s_ns=kafka_k8s_ns,
                                        kafka_tls_protocol=kafka_tls_protocol,
                                    )
                                else:
                                    self.create_kafka_receiver_starter(
                                        kafka_consumer_name_part,
                                        "consumer",
                                        start_json_state_options=start_json_state_consumer_options,
                                        kafka_options=kafka_consumer_options,
                                        stream_name=stream_name,
                                        kafka_producer_name=kafka_producer_name,
                                        kafka_receiver_name=kafka_receiver_name,
                                        receiver_name=receiver["name"],
                                        kafka_k8s_svc=kafka_k8s_svc,
                                        kafka_k8s_ns=kafka_k8s_ns,
                                        kafka_tls_protocol=kafka_tls_protocol,
                                    )

                        if self.receiver_starter_exists(kafka_consumer_name):
                            self.update_kafka_receiver_starter(
                                kafka_consumer_name,
                                "consumer",
                                start_json_state_options=start_json_state_consumer_options,
                                kafka_options=kafka_consumer_options,
                                stream_name=stream_name,
                                kafka_producer_name=kafka_producer_name,
                                kafka_receiver_name=kafka_receiver_name,
                                receiver_name=receiver["name"],
                                kafka_k8s_svc=kafka_k8s_svc,
                                kafka_k8s_ns=kafka_k8s_ns,
                                kafka_tls_protocol=kafka_tls_protocol,
                            )
                        else:
                            self.create_kafka_receiver_starter(
                                kafka_consumer_name,
                                "consumer",
                                start_json_state_options=start_json_state_consumer_options,
                                kafka_options=kafka_consumer_options,
                                stream_name=stream_name,
                                kafka_producer_name=kafka_producer_name,
                                kafka_receiver_name=kafka_receiver_name,
                                receiver_name=receiver["name"],
                                kafka_k8s_svc=kafka_k8s_svc,
                                kafka_k8s_ns=kafka_k8s_ns,
                                kafka_tls_protocol=kafka_tls_protocol,
                            )

        if stream_name and not self.stream_exists(stream_name):
            if kafka_receiver_name:
                self.create_stream(
                    stream_name,
                    "/le-mans/v2/resources/receivers/" + kafka_receiver_name,
                )
            else:
                self.create_stream(
                    stream_name, "/le-mans/v2/resources/receivers/" + receiver["name"]
                )

    def multiplexer_operations(self, kafka_receiver):
        buffer_time_millis = 100
        max_buffer_bytes_millis = 1000000
        queue_operation_timeout_millis = 10000
        queue_capacity = 10000
        headers_to_multiplex = []
        criteria_for_multiplexing = {}
        lemans_multiplexer_custom_headers = {}
        lemans_multiplexer_name = ""
        if "lemans_multiplexer_name" in kafka_receiver:
            lemans_multiplexer_name = kafka_receiver["lemans_multiplexer_name"]
            if "next_receiver_path" not in kafka_receiver:
                logging.error(
                    "Please specify nextReceiverPath in lemans hook configuration for %s",
                    lemans_multiplexer_name,
                )
            else:
                next_receiver_path = kafka_receiver["next_receiver_path"]
                if "criteria_for_multiplexing" in kafka_receiver:
                    criteria_for_multiplexing = kafka_receiver[
                        "criteria_for_multiplexing"
                    ]
                if "buffer_time_millis" in kafka_receiver:
                    buffer_time_millis = kafka_receiver["buffer_time_millis"]
                if "max_buffer_bytes_millis" in kafka_receiver:
                    max_buffer_bytes_millis = kafka_receiver["max_buffer_bytes_millis"]
                if "queue_operation_timeout_millis" in kafka_receiver:
                    queue_operation_timeout_millis = kafka_receiver[
                        "queue_operation_timeout_millis"
                    ]
                if "queue_capacity" in kafka_receiver:
                    queue_capacity = kafka_receiver["queue_capacity"]
                if "headers_to_multiplex" in kafka_receiver:
                    headers_to_multiplex = kafka_receiver["headers_to_multiplex"]
                if "lemans_multiplexer_custom_headers" in kafka_receiver:
                    lemans_multiplexer_custom_headers = kafka_receiver[
                        "lemans_multiplexer_custom_headers"
                    ]
                if self.receiver_starter_exists(lemans_multiplexer_name):
                    self.update_kafka_receiver_starter_multiplexer(
                        lemans_multiplexer_name,
                        "multiplexer",
                        nextReceiverPath=next_receiver_path,
                        bufferTimeMillis=buffer_time_millis,
                        maxBufferBytesMillis=max_buffer_bytes_millis,
                        queueOperationTimeoutMillis=queue_operation_timeout_millis,
                        queueCapacity=queue_capacity,
                        headersToMultiplex=headers_to_multiplex,
                        criteriaForMultiplexing=criteria_for_multiplexing,
                        customHeaders=lemans_multiplexer_custom_headers,
                    )
                else:
                    self.create_kafka_receiver_starter_multiplexer(
                        lemans_multiplexer_name,
                        "multiplexer",
                        nextReceiverPath=next_receiver_path,
                        bufferTimeMillis=buffer_time_millis,
                        maxBufferBytesMillis=max_buffer_bytes_millis,
                        queueOperationTimeoutMillis=queue_operation_timeout_millis,
                        queueCapacity=queue_capacity,
                        headersToMultiplex=headers_to_multiplex,
                        criteriaForMultiplexing=criteria_for_multiplexing,
                        customHeaders=lemans_multiplexer_custom_headers,
                    )

    def create_receiver(self, receiver, service_path):
        data = {
            "name": receiver["name"],
            "address": service_path,
            "useHttp2": receiver["useHttp2"],
            "operationTimeoutMicros": receiver["operationTimeoutMicros"],
        }
        if "customHeaders" in receiver:
            data["customHeaders"] = receiver["customHeaders"]
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/receivers"
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.post(
                url, json=data, headers={"Authorization": jwttoken}
            )
            sm_response.raise_for_status()
            logging.debug(
                "Created receiver %s for %s and the sm_url is %s",
                receiver["name"],
                service_path,
                url,
            )
            logging.debug("%s is the receiever sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.post(url, json=data)
            response.raise_for_status()
            logging.info(
                "Created receiver of %s for %s", receiver["name"], service_path
            )

    def create_route(self, route):
        data = {
            "selector": route["selector"],
            "allowedPath": route["allowedPath"],
            "allowedActions": route["allowedActions"],
            "receiverLink": route["receiverLink"],
        }
        if "retainAuthHeader" in route.keys():
            data["retainAuthHeader"] = route["retainAuthHeader"]
        if "retainUrlEncoding" in route.keys():
            data["retainUrlEncoding"] = route["retainUrlEncoding"]
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/routes"
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.post(
                url, json=data, headers={"Authorization": jwttoken}
            )
            sm_response.raise_for_status()
            logging.debug("Created route %s and the sm_url is %s", route["name"], url)
            logging.debug("%s is the route sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.post(url, json=data)
            response.raise_for_status()
            logging.info("Created route of %s", route["name"])

    def delete_route(self, route):
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/routes/"
            + route["name"]
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.delete(url, headers={"Authorization": jwttoken})
            sm_response.raise_for_status()
            logging.debug("deleted route %s", route["name"])
            logging.debug("%s is the delete sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.delete(url)
            response.raise_for_status()
            logging.info("deleted route %s", route["name"])

    def update_receiver(self, receiver, service_path):
        data = {
            "name": receiver["name"],
            "address": service_path,
            "useHttp2": receiver["useHttp2"],
            "operationTimeoutMicros": receiver["operationTimeoutMicros"],
        }
        if "customHeaders" in receiver:
            data["customHeaders"] = receiver["customHeaders"]
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/receivers/"
            + receiver["name"]
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.put(
                url, json=data, headers={"Authorization": jwttoken}
            )
            logging.debug(
                "Updated receiver %s for %s and the sm_url is %s",
                receiver["name"],
                service_path,
                url,
            )
            logging.debug("%s is the update reciever sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.put(url, json=data)
            logging.info(
                "Updated receiver of %s for %s", receiver["name"], service_path
            )

    def create_stream(self, name, receiver_link):
        data = {
            "name": name,
            "deliveryPolicy": "WAIT_ALL",
            "useHttp2": True,
            "receiverLinks": [receiver_link],
        }
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/streams"
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.post(
                url, json=data, headers={"Authorization": jwttoken}
            )
            sm_response.raise_for_status()
            logging.debug(
                "Created stream of %s for receiver link %s and the sm_url is %s",
                name,
                receiver_link,
                url,
            )
            logging.debug("%s is the stream creation sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.post(url, json=data)
            response.raise_for_status()
            logging.info(
                "Created stream of %s for receiver link %s", name, receiver_link
            )

    def create_sample_receiver_starter(
        self, receiver_starter_name="", stream_name="", sample_receiver_options=[]
    ):
        data = {
            "name": receiver_starter_name,
            "factoryLink": "/le-mans/sample-receivers",
            "startJsonState": self.get_sample_receiver_starter_json_state(
                stream_name=stream_name,
                receiver_starter_name=receiver_starter_name,
                sample_receiver_options=sample_receiver_options,
            ),
        }
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/receiver-starters"
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.post(
                url, json=data, headers={"Authorization": jwttoken}
            )
            sm_response.raise_for_status()
            logging.debug(
                "Created sample receiver of %s and the sm_url is %s",
                receiver_starter_name,
                url,
            )
            logging.debug("%s is the stream creation sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.post(url, json=data)
            response.raise_for_status()
            logging.info("Created sample receiver of %s", receiver_starter_name)

    def update_sample_receiver_starter(
        self, receiver_starter_name="", stream_name="", sample_receiver_options=[]
    ):
        update_required = self.check_change_in_start_json_data_sample_receiver(
            receiver_starter_name, stream_name, sample_receiver_options
        )
        if update_required:
            self.delete_resource("receiver-starters", receiver_starter_name)
            self.create_sample_receiver_starter(
                receiver_starter_name, stream_name, sample_receiver_options
            )
            logging.info("Updated sample receiver starter: %s", receiver_starter_name)

    def check_change_in_start_json_data_sample_receiver(
        self, receiver_starter_name="", stream_name="", sample_receiver_options=[]
    ):
        receiver_data = self.get_receiver_starter_json(receiver_starter_name)
        start_json_data = eval(receiver_data.get("startJsonState"))
        changes_start_json_data = eval(
            self.get_sample_receiver_starter_json_state(
                stream_name, receiver_starter_name, sample_receiver_options
            )
        )

        res1 = jsondiff.diff(start_json_data, changes_start_json_data)
        res2 = jsondiff.diff(changes_start_json_data, start_json_data)
        if res1:
            print("Difference in start_json_data found")
            print("Current is " + str(res2))
            print("Updates required is " + str(res1))
            return True

        return False

    def get_sample_receiver_starter_json_state(
        self,
        stream_name="",
        receiver_starter_name="",
        sample_receiver_options=[],
        return_json=False,
    ):
        start_json_state_dict = {
            "topic": stream_name,
            "contentType": "application/json",
            "receiverLink": "/le-mans/v2/resources/receivers/" + receiver_starter_name,
        }

        if len(sample_receiver_options) != 0:
            for key, value in sample_receiver_options.items():
                if key in start_json_state_dict:
                    start_json_state_dict[key] = value
                else:
                    start_json_state_dict.update({key: value})

        if return_json:
            start_json_state = json.dumps(start_json_state_dict)
        else:
            start_json_state = json.dumps(
                start_json_state_dict, separators=(",", ":")
            ).replace('"', "'")

        return start_json_state

    def create_kafka_receiver_starter(
        self,
        name,
        type,
        start_json_state_options=[],
        kafka_options=[],
        stream_name="",
        kafka_producer_name="",
        kafka_receiver_name="",
        receiver_name="",
        kafka_k8s_svc=KAFKA_SERVICE,
        kafka_k8s_ns="",
        kafka_tls_protocol=KAFKA_SECURITY_PROTOCOL,
    ):
        logging.info("Creating receiver starter type %s : %s", type, name)
        start_json_state = self.get_start_json_state(
            type=type,
            start_json_state_options=start_json_state_options,
            kafka_options=kafka_options,
            stream_name=stream_name,
            kafka_producer_name=kafka_producer_name,
            kafka_receiver_name=kafka_receiver_name,
            receiver_name=receiver_name,
            kafka_k8s_svc=kafka_k8s_svc,
            kafka_k8s_ns=kafka_k8s_ns,
            kafka_tls_protocol=kafka_tls_protocol,
        )
        if type == "producer":
            factory_link = "/le-mans/receivers/kafka-producer"
            startPriority = 1
        elif type == "consumer":
            factory_link = "/le-mans/consumers/kafka"
            startPriority = 0
        else:
            raise Exception

        data = {
            "name": name,
            "factoryLink": factory_link,
            "startJsonState": start_json_state,
            "startPriority": startPriority,
        }
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/receiver-starters"
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.post(
                url, json=data, headers={"Authorization": jwttoken}
            )
            sm_response.raise_for_status()
            logging.debug("Created receiver starter type : %s , name : %s ", type, name)
            logging.debug(
                "%s is the reciever-starter creation sm_response ", sm_response
            )
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.post(url, json=data)
            response.raise_for_status()
            logging.info("Created receiver starter type : %s , name : %s ", type, name)

    def create_egress_receiver_with_routes_and_consumer(
        self, stream_name="", receiver=[], kafka_receiver={}
    ):
        if not self.receiver_exists(receiver["name"]):
            self.create_receiver(receiver, receiver["receiver_address"])
        else:
            self.update_receiver(receiver, receiver["receiver_address"])
        if "routes" in receiver.keys():
            for route in receiver["routes"]:
                route["receiverLink"] = (
                    "/le-mans/v2/resources/receivers/" + receiver["name"]
                )
                route_var = self.route_exists_unchanged(route["name"], route)
                if route_var == "no route":
                    self.create_route(route)
                elif route_var == "changed route":
                    self.delete_route(route)
                    self.create_route(route)

            # Handles consumption from different products
        if "kafka_consumer_name" in receiver.keys():
            kafka_k8s_svc = KAFKA_SERVICE
            kafka_tls_protocol = KAFKA_SECURITY_PROTOCOL
            kafka_k8s_ns = KAFKA_NAMESPACE
            if "kafka_k8s_svc" in receiver.keys():
                kafka_k8s_svc = receiver["kafka_k8s_svc"]
            logging.info("we have one egress_receiver and one consumer")
            kafka_consumer_name = receiver["kafka_consumer_name"]
            if ("kafka_producer_name" in receiver.keys()) or (
                "kafka_producer_name" in kafka_receiver.keys()
            ):
                if "kafka_producer_name" not in kafka_receiver.keys():
                    kafka_producer_name = receiver["kafka_producer_name"]
                else:
                    kafka_producer_name = kafka_receiver["kafka_producer_name"]
                no_of_consumers = 1
                if "no_of_consumers" in receiver:
                    no_of_consumers = int(receiver["no_of_consumers"])
                if "start_json_state_consumer_options" in receiver:
                    start_json_state_consumer_options = receiver[
                        "start_json_state_consumer_options"
                    ]
                else:
                    start_json_state_consumer_options = ""
                if "kafka_consumer_options" in receiver:
                    kafka_consumer_options = receiver["kafka_consumer_options"]
                    if "group.id" not in kafka_consumer_options:
                        logging.error(
                            "Please specify group.id in kafka_consumer_options in lemans hook configuration for %s",
                            kafka_consumer_name,
                        )
                if "kafka_tls_protocol" in receiver:
                    kafka_tls_protocol = receiver["kafka_tls_protocol"]
                kafka_receiver_name = receiver["name"]
                for x in range(0, no_of_consumers):
                    if no_of_consumers != 1:
                        kafka_consumer_name_str = kafka_consumer_name + "-" + str(x)
                    else:
                        kafka_consumer_name_str = kafka_consumer_name
                    if self.receiver_starter_exists(kafka_consumer_name_str):
                        self.update_kafka_receiver_starter(
                            kafka_consumer_name_str,
                            "consumer",
                            start_json_state_options=start_json_state_consumer_options,
                            kafka_options=kafka_consumer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            receiver_name=receiver["name"],
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
                    else:
                        self.create_kafka_receiver_starter(
                            kafka_consumer_name_str,
                            "consumer",
                            start_json_state_options=start_json_state_consumer_options,
                            kafka_options=kafka_consumer_options,
                            stream_name=stream_name,
                            kafka_producer_name=kafka_producer_name,
                            kafka_receiver_name=kafka_receiver_name,
                            receiver_name=receiver["name"],
                            kafka_k8s_svc=kafka_k8s_svc,
                            kafka_k8s_ns=kafka_k8s_ns,
                            kafka_tls_protocol=kafka_tls_protocol,
                        )
            else:
                logging.error(
                    "Please specify kafka_producer_name for this consumer: %s",
                    kafka_consumer_name,
                )

    def create_kafka_receiver_starter_multiplexer(
        self,
        name,
        type,
        nextReceiverPath,
        bufferTimeMillis,
        maxBufferBytesMillis,
        queueOperationTimeoutMillis,
        queueCapacity,
        headersToMultiplex,
        criteriaForMultiplexing,
        customHeaders,
    ):
        if not nextReceiverPath:
            logging.info(
                "For creating multiplexer, Please add the value for nextReceiverPath as it is a mandatory field"
            )
            raise Exception
        logging.info("Creating receiver starter type %s : %s", type, name)
        start_json_state_multiplexer = self.get_start_json_state_multiplexer(
            type=type,
            nextReceiverPath=nextReceiverPath,
            bufferTimeMillis=bufferTimeMillis,
            maxBufferBytesMillis=maxBufferBytesMillis,
            queueOperationTimeoutMillis=queueOperationTimeoutMillis,
            queueCapacity=queueCapacity,
            headersToMultiplex=headersToMultiplex,
            criteriaForMultiplexing=criteriaForMultiplexing,
            customHeaders=customHeaders,
        )
        logging.info(
            "start_json_state_multiplexer : {0}".format(start_json_state_multiplexer)
        )
        if type == "multiplexer":
            factory_link = "/le-mans/receivers/multiplexer"
            startPriority = 0
        else:
            raise Exception

        data = {
            "name": name,
            "factoryLink": factory_link,
            "startJsonState": start_json_state_multiplexer,
            "startPriority": startPriority,
        }
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/receiver-starters"
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.post(
                url, json=data, headers={"Authorization": jwttoken}
            )
            sm_response.raise_for_status()
            logging.info("Created receiver starter type : %s , name : %s", type, name)
            logging.debug(
                "%s is the reciever-starter creation sm_response ", sm_response
            )
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.post(url, json=data)
            response.raise_for_status()
            logging.info("Created receiver starter type : %s , name : %s", type, name)

    def update_kafka_receiver_starter_multiplexer(
        self,
        lemans_multiplexer_name,
        type,
        nextReceiverPath,
        bufferTimeMillis,
        maxBufferBytesMillis,
        queueOperationTimeoutMillis,
        queueCapacity,
        headersToMultiplex,
        criteriaForMultiplexing,
        customHeaders,
    ):
        update_required = self.check_change_in_start_json_data_multiplexer(
            lemans_multiplexer_name,
            type,
            nextReceiverPath,
            bufferTimeMillis,
            maxBufferBytesMillis,
            queueOperationTimeoutMillis,
            queueCapacity,
            headersToMultiplex,
            criteriaForMultiplexing,
            customHeaders,
        )
        if update_required:
            # receiver-starter does not support PATCH
            # We cannot POST again to create a receiver-starter that already exists.
            # So we will DELETE and then POST.
            logging.info(
                "deleting receiver-starters type : %s , name : %s",
                type,
                lemans_multiplexer_name,
            )
            self.delete_resource("receiver-starters", lemans_multiplexer_name)
            self.create_kafka_receiver_starter_multiplexer(
                lemans_multiplexer_name,
                type,
                nextReceiverPath=nextReceiverPath,
                bufferTimeMillis=bufferTimeMillis,
                maxBufferBytesMillis=maxBufferBytesMillis,
                queueOperationTimeoutMillis=queueOperationTimeoutMillis,
                queueCapacity=queueCapacity,
                headersToMultiplex=headersToMultiplex,
                criteriaForMultiplexing=criteriaForMultiplexing,
                customHeaders=customHeaders,
            )
            logging.info(
                "Updated kafka receiver starter type : %s , name : %s",
                type,
                lemans_multiplexer_name,
            )

    def check_change_in_start_json_data_multiplexer(
        self,
        lemans_multiplexer_name,
        type,
        nextReceiverPath,
        bufferTimeMillis,
        maxBufferBytesMillis,
        queueOperationTimeoutMillis,
        queueCapacity,
        headersToMultiplex,
        criteriaForMultiplexing,
        customHeaders,
    ):
        receiver_data = self.get_receiver_starter_json(lemans_multiplexer_name)
        start_json_data = eval(receiver_data.get("startJsonState"))
        changes_start_json_data = eval(
            self.get_start_json_state_multiplexer(
                type=type,
                nextReceiverPath=nextReceiverPath,
                bufferTimeMillis=bufferTimeMillis,
                maxBufferBytesMillis=maxBufferBytesMillis,
                queueOperationTimeoutMillis=queueOperationTimeoutMillis,
                queueCapacity=queueCapacity,
                headersToMultiplex=headersToMultiplex,
                criteriaForMultiplexing=criteriaForMultiplexing,
                customHeaders=customHeaders,
            )
        )
        res1 = jsondiff.diff(start_json_data, changes_start_json_data)
        res2 = jsondiff.diff(changes_start_json_data, start_json_data)
        if res1:
            print("Difference in start_json_data found")
            print("Current is " + str(res2))
            print("Updates required is " + str(res1))
            return True
        return False

    def update_kafka_receiver_starter(
        self,
        receiver_starter_name,
        type,
        start_json_state_options=[],
        kafka_options=[],
        stream_name="",
        kafka_producer_name="",
        kafka_receiver_name="",
        receiver_name="",
        kafka_k8s_svc=KAFKA_SERVICE,
        kafka_k8s_ns="",
        kafka_tls_protocol=KAFKA_SECURITY_PROTOCOL,
    ):
        update_required = self.check_change_in_start_json_data(
            receiver_starter_name,
            type,
            start_json_state_options,
            kafka_options,
            stream_name,
            kafka_producer_name,
            kafka_receiver_name,
            receiver_name,
            kafka_k8s_svc,
            kafka_k8s_ns,
            kafka_tls_protocol,
        )
        if update_required:
            # receiver-starter does not support PATCH
            # We cannot POST again to create a receiver-starter that already exists.
            # So we will DELETE and then POST.

            self.delete_resource("receiver-starters", receiver_starter_name)
            if type == "producer":
                factory_link = "/le-mans/receivers/kafka-producer"
                startPriority = 1
            elif type == "consumer":
                factory_link = "/le-mans/consumers/kafka"
                startPriority = 0
            self.create_kafka_receiver_starter(
                receiver_starter_name,
                type,
                start_json_state_options,
                kafka_options,
                stream_name=stream_name,
                kafka_producer_name=kafka_producer_name,
                kafka_receiver_name=kafka_receiver_name,
                receiver_name=receiver_name,
                kafka_k8s_svc=kafka_k8s_svc,
                kafka_k8s_ns=kafka_k8s_ns,
                kafka_tls_protocol=kafka_tls_protocol,
            )
            logging.info(
                "Updated kafka receiver starter type : %s , name : %s",
                type,
                receiver_starter_name,
            )

    def check_change_in_start_json_data(
        self,
        receiver_starter_name,
        type,
        start_json_state_options,
        kafka_options,
        stream_name,
        kafka_producer_name,
        kafka_receiver_name,
        receiver_name,
        kafka_k8s_svc=KAFKA_SERVICE,
        kafka_k8s_ns="",
        kafka_tls_protocol=KAFKA_SECURITY_PROTOCOL,
    ):
        receiver_data = self.get_receiver_starter_json(receiver_starter_name)
        start_json_data = eval(receiver_data.get("startJsonState"))
        kafka_properties_data = start_json_data.get("kafkaProperties")
        changes_start_json_data = eval(
            self.get_start_json_state(
                type=type,
                start_json_state_options=start_json_state_options,
                kafka_options=kafka_options,
                stream_name=stream_name,
                kafka_producer_name=kafka_producer_name,
                kafka_receiver_name=kafka_receiver_name,
                receiver_name=receiver_name,
                return_json=True,
                kafka_k8s_svc=kafka_k8s_svc,
                kafka_k8s_ns=kafka_k8s_ns,
                kafka_tls_protocol=kafka_tls_protocol,
            )
        )
        changes_kafka_properties = changes_start_json_data.get("kafkaProperties")
        k1 = jsondiff.diff(kafka_properties_data, changes_kafka_properties)
        k2 = jsondiff.diff(changes_kafka_properties, kafka_properties_data)
        if k1:
            print("Difference in kafka properties found")
            print("Current is " + str(k2))
            print("Updates in kafka required properties is " + str(k1))
            return True

        del start_json_data["kafkaProperties"]
        del changes_start_json_data["kafkaProperties"]
        res1 = jsondiff.diff(start_json_data, changes_start_json_data)
        res2 = jsondiff.diff(changes_start_json_data, start_json_data)
        if res1:
            print("Difference in start_json_data found")
            print("Current is " + str(res2))
            print("Updates required is " + str(res1))
            return True

        return False

    def get_resource(self, resource_type, resource_name):
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/"
            + resource_type
            + "/"
            + resource_name
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.get(url, headers={"Authorization": jwttoken})
            logging.debug("the sm_url for get_resource is %s", url)
            logging.debug("%s is the get_resource sm_response ", sm_response)
            return sm_response.text
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.get(url)
            return response.text

    def get_resource_json(self, resource_type, resource_name):
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/"
            + resource_type
            + "/"
            + resource_name
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.get(url, headers={"Authorization": jwttoken})
            logging.debug("the sm_url for get_resource_json is %s", url)
            logging.debug("%s is the get_resource_json sm_response ", sm_response)
            return json.loads(sm_response.text)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.get(url)
            return json.loads(response.text)

    def get_receiver_starter_json(self, name):
        return self.get_resource_json("receiver-starters", name)

    def delete_resource(self, resource_type, resource_name):
        url = (
            "http://"
            + self.lemans_hostname
            + ":"
            + LEMANS_PORT
            + "/le-mans/v2/resources/"
            + resource_type
            + "/"
            + resource_name
        )
        if SM_EBALED:
            sm_result = requests.Session()
            sm_response = sm_result.delete(url, headers={"Authorization": jwttoken})
            sm_response.raise_for_status()
            logging.debug("the sm_url for delete_resource is %s", url)
            logging.debug("%s is the delete_resource sm_response ", sm_response)
        else:
            result = requests.Session()
            result.mount("http://", self.lemans_resources_adapter)
            response = result.delete(url)
            response.raise_for_status()

    def get_stream(self, name):
        return self.get_resource("streams", name)

    def delete_receiver(self, name):
        self.delete_resource("receivers", name)

    def get_receiver(self, name):
        return self.get_resource("receivers", name)

    def get_route(self, name):
        return self.get_resource("routes", name)

    def stream_exists(self, name):
        response = self.get_stream(name)
        return "link" in response

    def route_exists_unchanged(self, name, existing_route):
        route = self.get_route(name)
        route = json.loads(route)
        if not "link" in route:
            return "no route"
        if (
            route["selector"] == existing_route["selector"]
            and route["allowedActions"] == existing_route["allowedActions"]
            and route["allowedPath"] == existing_route["allowedPath"]
            and route["receiverLink"] == existing_route["receiverLink"]
        ):
            if ("retainAuthHeader" in existing_route.keys()) and (
                route["retainAuthHeader"] != existing_route["retainAuthHeader"]
            ):
                return "changed route"
            if ("retainUrlEncoding" in existing_route.keys()) and (
                route["retainUrlEncoding"] != existing_route["retainUrlEncoding"]
            ):
                return "changed route"
            return "unchanged route"
        return "changed route"

    def receiver_exists(self, name):
        receiver = self.get_receiver(name)
        return "link" in receiver

    def delete_receiver_starter(self, name):
        self.delete_resource("receiver-starters", name)

    def get_receiver_starter(self, name):
        return self.get_resource("receiver-starters", name)

    def receiver_starter_exists(self, name):
        response = self.get_receiver_starter(name)
        return "link" in response

    def get_start_json_state(
        self,
        type,
        start_json_state_options=[],
        kafka_options=[],
        stream_name="",
        kafka_producer_name="",
        kafka_receiver_name="",
        receiver_name="",
        return_json=False,
        kafka_k8s_svc=KAFKA_SERVICE,
        kafka_k8s_ns="",
        kafka_tls_protocol=KAFKA_SECURITY_PROTOCOL,
    ):
        start_json_state_dict = {}
        kafka_hostname = get_service_hostname(kafka_k8s_ns, kafka_k8s_svc)
        if type == "producer":
            start_json_state_dict = {
                "topicName": stream_name,
                "numberOfPartitions": KAFKA_NUMBER_OF_PARTITIONS,
                "replicationFactor": KAFKA_REPLICAS,
                "retentionPeriod": KAFKA_RETENTION_PERIOD,
                "cleanupPolicy": KAFKA_CLEANUP_POLICY,
            }
            if kafka_tls_protocol == "SSL":
                start_json_state_dict["kafkaProperties"] = {
                    "bootstrap.servers": kafka_hostname + ":9092",
                    "acks": KAFKA_ACKS,
                    "retries": KAFKA_RETRIES,
                    "batch.size": KAFKA_BATCH_SIZE,
                    "max.block.ms": KAFKA_MAX_BLOCK_MS,
                    "max.request.size": KAFKA_MAX_REQUEST_SIZE,
                    "linger.ms": KAFKA_LINGER_MS,
                    "buffer.memory": KAFKA_BUFFER_MEMORY,
                    "key.serializer": KAFKA_KEY_SERIALIZER,
                    "value.serializer": KAFKA_VALUE_SERIALIZER,
                    "partitioner.class": KAFKA_PARTITIONER_CLASS,
                    "security.protocol": "SSL",
                    "ssl.keystore.location": "/opt/le-mans/kafka/ssl/lemans.diploma.keystore.jks",
                    "ssl.keystore.password": KAFKA_SSL_KEYSTORE_PASS,
                    "ssl.key.password": KAFKA_SSL_KEY_PASS,
                    "ssl.truststore.location": "/opt/le-mans/kafka/ssl/lemans.diploma.truststore.jks",
                    "ssl.truststore.password": KAFKA_SSL_TRUSTSTORE_PASS,
                    "compression.type": "zstd",
                    "ssl.endpoint.identification.algorithm": "",
                }
            elif kafka_tls_protocol == "SSL_TO_PLAINTEXT":
                start_json_state_dict["kafkaProperties"] = {
                    "bootstrap.servers": kafka_hostname + ":9093",
                    "acks": KAFKA_ACKS,
                    "retries": KAFKA_RETRIES,
                    "batch.size": KAFKA_BATCH_SIZE,
                    "max.block.ms": KAFKA_MAX_BLOCK_MS,
                    "max.request.size": KAFKA_MAX_REQUEST_SIZE,
                    "linger.ms": KAFKA_LINGER_MS,
                    "buffer.memory": KAFKA_BUFFER_MEMORY,
                    "key.serializer": KAFKA_KEY_SERIALIZER,
                    "value.serializer": KAFKA_VALUE_SERIALIZER,
                    "partitioner.class": KAFKA_PARTITIONER_CLASS,
                    "security.protocol": "PLAINTEXT",
                    "compression.type": "zstd",
                }
            else:
                start_json_state_dict["kafkaProperties"] = {
                    "bootstrap.servers": kafka_hostname + ":9092",
                    "acks": KAFKA_ACKS,
                    "retries": KAFKA_RETRIES,
                    "batch.size": KAFKA_BATCH_SIZE,
                    "max.block.ms": KAFKA_MAX_BLOCK_MS,
                    "max.request.size": KAFKA_MAX_REQUEST_SIZE,
                    "linger.ms": KAFKA_LINGER_MS,
                    "buffer.memory": KAFKA_BUFFER_MEMORY,
                    "key.serializer": KAFKA_KEY_SERIALIZER,
                    "value.serializer": KAFKA_VALUE_SERIALIZER,
                    "partitioner.class": KAFKA_PARTITIONER_CLASS,
                    "security.protocol": "PLAINTEXT",
                    "compression.type": "zstd",
                }
        elif type == "consumer":
            start_json_state_dict = {
                "topic": stream_name,
                "retryTopic": stream_name,
                "kafkaProducerPath": "/le-mans/receivers/kafka-producer/"
                + kafka_producer_name,
                "maxRetryLimit": 3,
                "contentType": "application/json",
                "receiverLink": "/le-mans/v2/resources/receivers/" + receiver_name,
            }
            if kafka_tls_protocol == "SSL":
                start_json_state_dict["kafkaProperties"] = {
                    "bootstrap.servers": kafka_hostname + ":9092",
                    "group.id": "le-mans",
                    "fetch.min.bytes": 1,
                    "maxRetryLimit": 3,
                    "max.poll.records": 5,
                    "key.deserializer": KAFKA_KEY_DESERIALIZER,
                    "value.deserializer": KAFKA_VALUE_DESERIALIZER,
                    "ssl.key.password": KAFKA_SSL_KEY_PASS,
                    "auto.offset.reset": "latest",
                    "security.protocol": "SSL",
                    "ssl.keystore.location": "/opt/le-mans/kafka/ssl/lemans.diploma.keystore.jks",
                    "ssl.keystore.password": KAFKA_SSL_KEYSTORE_PASS,
                    "ssl.truststore.location": "/opt/le-mans/kafka/ssl/lemans.diploma.truststore.jks",
                    "ssl.truststore.password": KAFKA_SSL_TRUSTSTORE_PASS,
                    "ssl.endpoint.identification.algorithm": "",
                }
            elif kafka_tls_protocol == "SSL_TO_PLAINTEXT":
                start_json_state_dict["kafkaProperties"] = {
                    "bootstrap.servers": kafka_hostname + ":9093",
                    "group.id": "le-mans",
                    "fetch.min.bytes": 1,
                    "maxRetryLimit": 3,
                    "max.poll.records": 5,
                    "key.deserializer": KAFKA_KEY_DESERIALIZER,
                    "value.deserializer": KAFKA_VALUE_DESERIALIZER,
                    "auto.offset.reset": "latest",
                    "security.protocol": "PLAINTEXT",
                }
            else:
                start_json_state_dict["kafkaProperties"] = {
                    "bootstrap.servers": kafka_hostname + ":9092",
                    "group.id": "le-mans",
                    "fetch.min.bytes": 1,
                    "maxRetryLimit": 3,
                    "max.poll.records": 5,
                    "key.deserializer": KAFKA_KEY_DESERIALIZER,
                    "value.deserializer": KAFKA_VALUE_DESERIALIZER,
                    "auto.offset.reset": "latest",
                    "security.protocol": "PLAINTEXT",
                }
        else:
            logging.error("The type should be producer or consumer")
            raise Exception

        if len(start_json_state_options) != 0:
            for key, value in start_json_state_options.items():
                if key in start_json_state_dict:
                    start_json_state_dict[key] = value
                else:
                    start_json_state_dict.update({key: value})

        if len(kafka_options) != 0:
            for key, value in kafka_options.items():
                if key in start_json_state_dict["kafkaProperties"]:
                    start_json_state_dict["kafkaProperties"][key] = value
                else:
                    start_json_state_dict["kafkaProperties"].update({key: value})

        if (
            "partitioner.class" in start_json_state_dict["kafkaProperties"]
            and len(start_json_state_dict["kafkaProperties"]["partitioner.class"]) == 0
        ):
            del start_json_state_dict["kafkaProperties"]["partitioner.class"]

        if return_json:
            start_json_state = json.dumps(start_json_state_dict)
        else:
            start_json_state = json.dumps(
                start_json_state_dict, separators=(",", ":")
            ).replace('"', "'")

        return start_json_state

    def get_start_json_state_multiplexer(
        self,
        type,
        nextReceiverPath,
        bufferTimeMillis,
        maxBufferBytesMillis,
        queueOperationTimeoutMillis,
        queueCapacity,
        headersToMultiplex,
        criteriaForMultiplexing,
        customHeaders,
        return_json=False,
    ):
        start_json_state_dict = {
            "nextReceiverPath": nextReceiverPath,
            "bufferTimeMillis": bufferTimeMillis,
            "maxBufferBytesMillis": maxBufferBytesMillis,
            "queueOperationTimeoutMillis": queueOperationTimeoutMillis,
            "queueCapacity": queueCapacity,
            "headersToMultiplex": headersToMultiplex,
            "customHeaders": {},
            "criteriaForMultiplexing": {},
        }

        print("the start json state dict is ", start_json_state_dict)
        print("customHeaders are : ", customHeaders)
        if len(customHeaders) != 0:
            for key, value in customHeaders.items():
                if key in start_json_state_dict["customHeaders"]:
                    start_json_state_dict["customHeaders"][key] = value
                else:
                    start_json_state_dict["customHeaders"].update({key: value})

        if len(criteriaForMultiplexing) != 0:
            for key, value in criteriaForMultiplexing.items():
                if key in start_json_state_dict["criteriaForMultiplexing"]:
                    start_json_state_dict["criteriaForMultiplexing"][key] = value
                else:
                    start_json_state_dict["criteriaForMultiplexing"].update(
                        {key: value}
                    )

        if return_json:
            start_json_state_multiplexer = json.dumps(start_json_state_dict)
        else:
            start_json_state_multiplexer = json.dumps(
                start_json_state_dict, separators=(",", ":")
            ).replace('"', "'")
        logging.info(
            "multiplexer final start json state : {0}".format(
                start_json_state_multiplexer
            )
        )
        return start_json_state_multiplexer


class AuthenticatingAdapter(requests.adapters.HTTPAdapter):
    """Injects a csp token and proxies the request"""

    def __init__(self, token_cb):
        self.token = None
        self.token_cb = token_cb
        self.session = requests.Session()
        super().__init__()

    def send(self, request, **kwargs):
        if self.token is None and self.token_cb:
            self.token = self.token_cb()
            logging.info("Injected token=%s", self.token)
        request.headers["x-xenon-auth-token"] = self.token

        if request.method in ("POST", "PUT", "PATCH"):
            kwargs["data"] = request.body

        debug_info = {
            "url": request.url,
            "headers": request.headers,
            "method": request.method,
            "data": request.body,
        }
        logging.info("debug info  : " + pprint.pformat(debug_info))

        req_executor = getattr(self.session, request.method.lower())
        return req_executor(request.url, headers=request.headers, **kwargs)


def session(adapter):
    """Construct a requests session that will submit requests to cerebro.

    :param adapter: A KubernetesProxyAdapter for submitting requests.
    :return: A requests.Session() where requests to http://anyhost/ will be
        submitted via the k8s proxy to the cerebro service.
    """
    result = requests.Session()
    result.mount("http://", adapter)
    return result


def get_service_hostname(namespace, service):
    return service + "." + namespace + ".svc.cluster.local"
